import datetime
import json
import time
from typing import Optional

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2 as FeatureSetApi
import ai.h2o.featurestore.api.v1.OnlineApi_pb2 as OnlineApi
import ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2 as TimeToLiveApi

from .. import interactive_console
from ..access_type import AccessType
from ..collections.ingest_history import IngestHistory
from ..collections.scheduled_tasks import ScheduledTasks
from ..commons.case_insensitive_dict import CaseInsensitiveDict
from ..credentials import CredentialsHelper
from ..data_source_wrappers import SparkDataFrame
from ..entities.backfill_option import BackfillOption
from ..entities.feature import Feature
from ..job_info import JobInfo
from ..retrieve_holder import RetrieveHolder
from ..utils import Utils
from .feature_set_ref import FeatureSetRef
from .feature_set_schema import FeatureSetSchema
from .ingest_job import IngestJob
from .materialization_online_job import MaterializationOnlineJob
from .recommendation import Recommendation
from .user import User


class FeatureSet:
    def __init__(self, stub, feature_set):
        fs = pb.FeatureSet()
        fs.CopyFrom(feature_set)
        self._feature_set = fs
        self._stub = stub

    @property
    def id(self):
        return self._feature_set.id

    @property
    def project(self):
        return self._feature_set.project

    @property
    def feature_set_name(self):
        return self._feature_set.feature_set_name

    @property
    def version(self) -> str:
        return self._feature_set.version

    @property
    def major_version(self) -> int:
        return int(self.version.split(".")[0])

    @property
    def version_change(self):
        return self._feature_set.version_change

    @property
    def time_travel_column(self):
        return self._feature_set.time_travel_column

    @property
    def partition_by(self):
        return self._feature_set.partition_by

    @property
    def time_travel_column_format(self):
        return self._feature_set.time_travel_column_format

    @property
    def feature_set_type(self):
        return FeatureSetApi.FeatureSetType.Name(self._feature_set.feature_set_type)

    @feature_set_type.setter
    def feature_set_type(self, value):
        valid_values = FeatureSetApi.FeatureSetType.keys()
        if value.upper() in valid_values:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._feature_set.id,
                feature_set_version=self._feature_set.version,
                type=FeatureSetApi.FeatureSetType.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TYPE],
            )
            self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception("Invalid feature set type. Supported values are: " + ", ".join(map(str, valid_values)))

    @property
    def description(self):
        return self._feature_set.description

    @description.setter
    def description(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            description=value,
            fields_to_update=[pb.FEATURE_SET_DESCRIPTION],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def owner(self):
        print('Owner field will be removed in 0.14.0. Please use "owners" instead.')
        return User(self._feature_set.owner)

    @owner.setter
    def owner(self, email):
        print('Owner field will be removed in 0.14.0. Please use "owners" instead.')
        request = pb.GetUserByMailRequest()
        request.email = email
        response = self._stub.GetUserByMail(request)
        user = response.user
        update_request = pb.FeatureSetUserFieldUpdateRequest()
        update_request.new_value.CopyFrom(user)
        header = FeatureSetApi.FeatureSetHeader(
            project_id=self._feature_set.project_id,
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        update_request.header.CopyFrom(header)
        self._stub.UpdateFeatureSetOwner(update_request)
        self.refresh()

    @property
    def author(self):
        return User(self._feature_set.author)

    @property
    def created_date_time(self):
        return Utils.timestamp_to_string(self._feature_set.created_date_time)

    @property
    def last_update_date_time(self):
        return Utils.timestamp_to_string(self._feature_set.last_update_date_time)

    @property
    def application_name(self):
        return self._feature_set.application_name

    @application_name.setter
    def application_name(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            application_name=value,
            fields_to_update=[pb.FEATURE_SET_APPLICATION_NAME],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def deprecated(self):
        return self._feature_set.deprecated

    @deprecated.setter
    def deprecated(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            deprecated=value,
            fields_to_update=[pb.FEATURE_SET_DEPRECATED],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def deprecated_date(self):
        return Utils.timestamp_to_string(self._feature_set.deprecated_date)

    @property
    def data_source_domains(self):
        return self._feature_set.data_source_domains

    @data_source_domains.setter
    def data_source_domains(self, value):
        if not isinstance(value, list):
            raise ValueError("data_source_domains accepts only list of strings as a value")

        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            fields_to_update=[pb.FEATURE_SET_DATA_SOURCE_DOMAINS],
        )
        update_request.data_source_domains.extend(value)
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def tags(self):
        return self._feature_set.tags

    @tags.setter
    def tags(self, value):
        if not isinstance(value, list):
            raise ValueError("tags accepts only list of strings as a value")

        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            fields_to_update=[pb.FEATURE_SET_TAGS],
        )
        update_request.tags.extend(value)
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def process_interval(self):
        return self._feature_set.process_interval

    @process_interval.setter
    def process_interval(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            process_interval=value,
            fields_to_update=[pb.FEATURE_SET_PROCESS_INTERVAL],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def process_interval_unit(self):
        return pb.ProcessIntervalUnit.Name(self._feature_set.process_interval_unit)

    @process_interval_unit.setter
    def process_interval_unit(self, value):
        valid_units = pb.ProcessIntervalUnit.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._feature_set.id,
                feature_set_version=self._feature_set.version,
                process_interval_unit=pb.ProcessIntervalUnit.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_PROCESS_INTERVAL_UNIT],
            )
            self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception("Invalid process interval unit. Supported values are: " + ", ".join(map(str, valid_units)))

    @property
    def flow(self):
        return self._feature_set.flow

    @flow.setter
    def flow(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            flow=value,
            fields_to_update=[pb.FEATURE_SET_FLOW],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def features(self):
        return CaseInsensitiveDict(
            {feature.name: Feature(self._stub, self, feature, feature.name) for feature in self._feature_set.features}
        )

    @property
    def primary_key(self):
        return self._feature_set.primary_key

    @primary_key.setter
    def primary_key(self, value):
        print(
            "Primary key setter will be removed in 0.14.0. Please use "
            + "createNewVersion with primaryKey argument instead."
        )
        update_request = pb.FeatureSetStringArrayFieldUpdateRequest()
        if isinstance(value, str):
            update_request.new_value.append(value)
        else:
            update_request.new_value.extend(value)
        header = FeatureSetApi.FeatureSetHeader(
            project_id=self._feature_set.project_id,
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        update_request.header.CopyFrom(header)
        self._stub.UpdateFeatureSetPrimaryKey(update_request)
        self.refresh()

    @property
    def statistics(self):
        return Statistics(self._feature_set)

    @property
    def time_to_live(self):
        return TimeToLive(self._stub, self)

    @property
    def special_data(self):
        return FeatureSetSpecialData(self._stub, self)

    @property
    def time_travel_scope(self):
        return FeatureSetScope(self._feature_set)

    @property
    def application_id(self):
        return self._feature_set.application_id

    @application_id.setter
    def application_id(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            application_id=value,
            fields_to_update=[pb.FEATURE_SET_APPLICATION_ID],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def feature_set_state(self):
        return self._feature_set.feature_set_state

    @feature_set_state.setter
    def feature_set_state(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            state=value,
            fields_to_update=[pb.FEATURE_SET_STATE],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def online(self):
        return Online(self._feature_set)

    @property
    def secret(self):
        return self._feature_set.secret

    @secret.setter
    def secret(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            secret=value,
            fields_to_update=[pb.FEATURE_SET_SECRET],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def custom_data(self):
        return self._feature_set.custom_data

    @property
    def feature_classifiers(self):
        return self._feature_set.feature_classifiers

    def is_derived(self):
        return self._feature_set.derived_from.HasField("transformation")

    def _reference(self) -> FeatureSetRef:
        return FeatureSetRef(self.id, self.major_version)

    @custom_data.setter
    def custom_data(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            custom_data=value,
            fields_to_update=[pb.FEATURE_SET_CUSTOM_DATA],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    def create_new_version(
        self,
        schema=None,
        affected_features=None,
        reason="",
        primary_key=None,
        backfill_options: Optional[BackfillOption] = None,
    ):
        if schema is None and affected_features is None and primary_key is None:
            raise ValueError(
                "At least one of schema, affected_features or primary_key must be defined. "
                + "All values are supported as well"
            )
        request = pb.CreateNewFeatureSetVersionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.reason = reason
        if schema:
            request.schema.extend(schema._to_proto_schema())
            if schema.is_derived():
                request.derived_from.CopyFrom(schema.derivation._to_proto())
        else:
            request.schema.extend(self.schema.get()._to_proto_schema())
        if affected_features:
            request.affected_features.extend(affected_features)
        if primary_key:
            request.primary_key.extend(primary_key)
        else:
            request.use_primary_key_from_previous_version = True
        if backfill_options:
            request.backfill_options.CopyFrom(backfill_options._to_proto(self._stub))
        response = self._stub.CreateNewFeatureSetVersion(request)
        if backfill_options:
            print(f"Backfill started with job id: '{response.job.job_id}'. Please use this job id to track progress.")
        return FeatureSet(self._stub, response.feature_set)

    def refresh(self):
        header = FeatureSetApi.FeatureSetHeader(
            project_id=self._feature_set.project_id,
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        request = pb.GetFeatureSetsLastMinorForCurrentMajorRequest()
        request.header.CopyFrom(header)
        self._feature_set = self._stub.GetFeatureSetsLastMinorForCurrentMajor(request).feature_set
        return self

    def delete(self, wait_for_completion=False):
        request = pb.DeleteFeatureSetRequest()
        request.feature_set.CopyFrom(self._feature_set)
        self._stub.DeleteFeatureSet(request)
        exists_request = pb.FeatureSetExistsRequest()
        exists_request.project_id = self._feature_set.project_id
        exists_request.feature_set_id = self._feature_set.id
        if wait_for_completion:
            while self._stub.FeatureSetExists(exists_request).exists:
                time.sleep(1)
                print("Waiting for feature set '{}' deletion".format(self._feature_set.feature_set_name))

    def add_owners(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Owner)

    def add_editors(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Editor)

    def add_consumers(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Consumer)

    def add_sensitive_consumers(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def add_viewers(self, user_emails):
        """Add additional viewer/viewers to a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.add_viewers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.Viewer)

    def remove_owners(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Owner)

    def remove_editors(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Editor)

    def remove_consumers(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Consumer)

    def remove_sensitive_consumers(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def remove_viewers(self, user_emails):
        """Remove viewer/viewers from a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.remove_viewers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.Viewer)

    def get_active_jobs(self, job_type=pb.JobType.Unknown):
        return self._get_jobs(True, job_type)

    def _get_jobs(self, active, job_type=pb.JobType.Unknown):
        from ..collections.jobs import Jobs  # Lazy import to avoid circular reference

        request = pb.ListJobsRequest(active=active)
        request.feature_set.CopyFrom(self._feature_set)
        request.job_type = job_type
        resp = self._stub.ListJobs(request)
        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]

    def _add_permissions(self, user_emails, permission):
        request = pb.FeatureSetPermissionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.AddFeatureSetPermission(request)
        return self

    def _remove_permissions(self, user_emails, permission):
        request = pb.FeatureSetPermissionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.RemoveFeatureSetPermission(request)
        return self

    def request_access(self, access_type, reason):
        request = pb.SubmitPendingPermissionRequest()
        request.resource_id = self._feature_set.id
        request.permission = AccessType.to_proto_permission(access_type)
        request.reason = reason
        response = self._stub.SubmitPendingFeatureSetPermission(request)
        return response.permission_id

    @property
    def current_permission(self):
        request = pb.GetActivePermissionRequest(resource_id=self._feature_set.id)
        response = self._stub.GetActiveFeatureSetPermission(request)
        return AccessType.from_proto_active_permission(response.permission)

    def ingest_async(
        self,
        source,
        credentials=None,
    ):
        if self.is_derived():
            raise Exception("Manual ingest is not allowed on derived feature set")
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            source._write_to_cache(self._stub)
            data_source = source._get_cache_location()
        else:
            data_source = get_raw_data_location(source)
        request = pb.StartIngestJobRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.data_source.CopyFrom(data_source)
        CredentialsHelper.set_credentials(request, data_source, credentials)
        job_id = self._stub.StartIngestJob(request)
        return IngestJob(self._stub, job_id)

    @interactive_console.record_stats
    def ingest(
        self,
        source,
        credentials=None,
    ):
        job = self.ingest_async(source, credentials)
        result = job.wait_for_result()
        self._feature_set = result._get_feature_set()
        return result

    def materialize_online_async(self):
        request = pb.StartMaterializationOnlineRequest()
        request.feature_set.CopyFrom(self._feature_set)
        job_id = self._stub.StartMaterializationOnlineJob(request)
        return MaterializationOnlineJob(self._stub, job_id)

    @interactive_console.record_stats
    def materialize_online(self):
        job = self.materialize_online_async()
        return job.wait_for_result()

    def ingest_online(self, rows):
        if self.is_derived():
            raise Exception("Manual ingest online is not allowed on derived feature set")

        if isinstance(rows, list):
            row_list = rows
        else:
            row_list = [rows]
        request = OnlineApi.OnlineIngestRequest()
        header = FeatureSetApi.FeatureSetHeader(
            project_id=self._feature_set.project_id,
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        request.header.CopyFrom(header)
        request.rows.extend(row_list)
        self._stub.OnlineIngest(request)

    def retrieve_online(self, *key) -> dict:
        request = OnlineApi.OnlineRetrieveRequest()
        header = FeatureSetApi.FeatureSetHeader(
            project_id=self._feature_set.project_id,
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        request.header.CopyFrom(header)
        request.key.extend(map(lambda x: str(x), key))
        json_row = self._stub.OnlineRetrieve(request).row
        return json.loads(json_row)

    def retrieve(self, start_date_time=None, end_date_time=None):
        return RetrieveHolder(self._stub, self._feature_set, start_date_time, end_date_time, "")

    def list_versions(self):
        request = pb.ListFeatureSetsVersionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        response = self._stub.ListFeatureSetVersions(request)
        return [VersionDescription(version) for version in response.versions]

    @property
    def schema(self):
        return FeatureSetSchema(self._stub, self)

    def ingest_history(self):
        return IngestHistory(self._stub, self._feature_set)

    def get_recommendations(self):
        request = pb.GetRecommendationRequest()
        request.feature_set.CopyFrom(self._feature_set)
        response = self._stub.GetRecommendations(request)
        return [Recommendation(self._stub, self, item) for item in response.matches]

    def schedule_ingest(self, name: str, source, schedule: str, description: str = "", credentials=None):
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            raise Exception("Scheduling Ingest with SparkDataFrame is not supported.")
        else:
            data_source = get_raw_data_location(source)

        client_timezone = datetime.datetime.utcnow().astimezone().tzname()

        request = pb.ScheduleTaskRequest(
            name=name,
            description=description,
            feature_set_id=self._feature_set.id,
            project_id=self._feature_set.project_id,
            source=data_source,
            schedule=schedule,
            feature_set_version=self._feature_set.version,
            crone_time_zone=client_timezone,
        )
        CredentialsHelper.set_credentials(request, data_source, credentials)
        scheduled_task = ScheduledTasks(self._stub, self)
        return scheduled_task.create_ingest_task(request)

    @property
    def schedule(self):
        return ScheduledTasks(self._stub, self)

    def get_preview(self):
        request = pb.GetFeatureSetPreviewRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        response = self._stub.GetFeatureSetPreview(request)
        if response.preview_url:
            json_response = Utils.fetch_preview_as_json_array(response.preview_url)
            return json_response
        else:
            print("No preview available. Job will be started to generate it.")
            job_id = self._start_preview_job()
            info = JobInfo(self._stub, job_id)
            while not self._get_job(job_id).done:
                info.show_progress()
                time.sleep(2)
            info.show_progress()
            return self.get_preview()

    def ingest_lazy(self, source, credentials=None):
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            raise Exception("Lazy Ingest with SparkDataFrame is not supported.")
        else:
            data_source = get_raw_data_location(source)
        request = pb.ScheduleTaskRequest(
            feature_set_id=self._feature_set.id,
            project_id=self._feature_set.project_id,
            source=data_source,
            feature_set_version=self._feature_set.version,
        )
        CredentialsHelper.set_credentials(request, data_source, credentials)
        scheduled_task = ScheduledTasks(self._stub, self)
        scheduled_task.create_lazy_ingest_task(request)
        print(f"Lazy ingest scheduled for feature set {self._feature_set.id}")

    @property
    def owners(self):
        request = pb.GetOwnersRequest(resource_id=self._feature_set.id)
        response = self._stub.GetFeatureSetOwners(request)
        return [User(owner) for owner in response.owners]

    def _start_preview_job(self):
        request = pb.PreviewRequest()
        request.feature_set.CopyFrom(self._feature_set)
        return self._stub.StartPreviewJob(request)

    def _get_job(self, job_id):
        return self._stub.GetJob(job_id)

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature_set)

    def __str__(self):
        return (
            f"Feature set name    : {self.feature_set_name} \n"
            f"Description         : {self.description} \n"
            f"Version             : {self.version} \n"
            f"Secret              : {self.secret} \n"
            f"Author                \n{Utils.output_indent_spacing(str(self.author), '      ')}"
            f"Project name        : {self.project} \n"
            f"Primary key         : {self.primary_key} \n"
            f"Feature set type    : {self.feature_set_type} \n"
            f"Created             : {self.created_date_time} \n"
            f"Last updated        : {self.last_update_date_time} \n"
            f"Sensitive           : {self.special_data._fs._feature_set.special_data.sensitive} \n"
            f"Time travel column  : {self.time_travel_column} \n"
            f"Features            : {self._custom_feature_fields()} \n"
            f"Tags                : {self.tags} \n"
            f"Partition by        : {self.partition_by} \n"
            f"Feature classifiers : {self.feature_classifiers} \n"
        )

    def _custom_feature_fields(self):
        tmp_dict = dict()
        for feature in self.features:
            tmp_dict.update({feature: self.features.get(feature).data_type})
        return json.dumps(tmp_dict, indent=5)


class VersionDescription:
    def __init__(self, version_description):
        self._version_description = version_description

    def __repr__(self):
        return Utils.pretty_print_proto(self._version_description)

    def __str__(self):
        return (
            f"Version           : {Utils.proto_to_dict(self._version_description).get('version')} \n"
            f"Version change    : {Utils.proto_to_dict(self._version_description).get('version_change')} \n"
            f"Created           : {Utils.proto_to_dict(self._version_description).get('created_date_time')} \n"
        )


class TimeToLive:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def ttl_offline(self):
        return self._fs._feature_set.time_to_live.ttl_offline

    @ttl_offline.setter
    def ttl_offline(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            time_to_live_offline_interval=value,
            fields_to_update=[pb.FEATURE_SET_TTL_OFFLINE],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def ttl_offline_interval(self):
        return TimeToLiveApi.Offline.TimeToLiveInterval.Name(self._fs._feature_set.time_to_live.ttl_offline_interval)

    @ttl_offline_interval.setter
    def ttl_offline_interval(self, value):
        valid_units = TimeToLiveApi.Offline.TimeToLiveInterval.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._fs._feature_set.id,
                feature_set_version=self._fs._feature_set.version,
                time_to_live_offline_interval_unit=TimeToLiveApi.Offline.TimeToLiveInterval.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TTL_OFFLINE_INTERVAL],
            )
            self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception(
                "Invalid offline time to live interval unit. Supported values are: " + ", ".join(map(str, valid_units))
            )

    @property
    def ttl_online(self):
        return self._fs._feature_set.time_to_live.ttl_online

    @ttl_online.setter
    def ttl_online(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            time_to_live_online_interval=value,
            fields_to_update=[pb.FEATURE_SET_TTL_ONLINE],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def ttl_online_interval(self):
        return TimeToLiveApi.Online.TimeToLiveInterval.Name(self._fs._feature_set.time_to_live.ttl_online_interval)

    @ttl_online_interval.setter
    def ttl_online_interval(self, value):
        valid_units = TimeToLiveApi.Online.TimeToLiveInterval.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._fs._feature_set.id,
                feature_set_version=self._fs._feature_set.version,
                time_to_live_online_interval_unit=TimeToLiveApi.Online.TimeToLiveInterval.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TTL_ONLINE_INTERVAL],
            )
            self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception(
                "Invalid online time to live interval unit. Supported values are: " + ", ".join(map(str, valid_units))
            )

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.time_to_live)

    def __str__(self):
        return (
            f"ttl offline           : {self.ttl_offline} \n"
            f"ttl online            : {self.ttl_online} \n"
            f"ttl offline interval  : {self.ttl_offline_interval} \n"
            f"ttl online interval   : {self.ttl_online_interval} \n"
        )


class FeatureSetScope:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._scope = self._feature_set.time_travel_scope

    @property
    def start_date_time(self):
        return Utils.timestamp_to_string(self._scope.start_date_time)

    @property
    def end_date_time(self):
        return Utils.timestamp_to_string(self._scope.end_date_time)

    def __repr__(self):
        return Utils.pretty_print_proto(self._scope)

    def __str__(self):
        return f"Start date & time : {self.start_date_time} \n" f"End date & time   : {self.end_date_time} \n"


class FeatureSetSpecialData:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def spi(self):
        return self._fs._feature_set.special_data.spi

    @property
    def pci(self):
        return self._fs._feature_set.special_data.pci

    @property
    def rpi(self):
        return self._fs._feature_set.special_data.rpi

    @property
    def demographic(self):
        return self._fs._feature_set.special_data.demographic

    @property
    def legal(self):
        return Legal(self._stub, self._fs)

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.special_data)

    def __str__(self):
        return (
            f"legal           \n{Utils.output_indent_spacing(str(self.legal), '    ')}"
            f"spi           : {self.spi} \n"
            f"pci           : {self.pci} \n"
            f"rpi           : {self.rpi} \n"
            f"demographic   : {self.demographic} \n"
            f"sensitive     : {Utils.proto_to_dict(self._fs._feature_set.special_data).get('sensitive')} \n"
        )


class Statistics:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._statistics = self._feature_set.statistics

    @property
    def data_latency(self):
        return self._statistics.data_latency

    def __repr__(self):
        return Utils.pretty_print_proto(self._statistics)

    def __str__(self):
        return f"Data latency  : {self.data_latency} \n"


class Legal:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def approved(self):
        return self._fs._feature_set.special_data.legal.approved

    @approved.setter
    def approved(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            legal_approved=value,
            fields_to_update=[pb.FEATURE_SET_SPECIAL_DATA_LEGAL_APPROVED],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def approved_date(self):
        return Utils.timestamp_to_string(self._fs._feature_set.special_data.legal.approved_date)

    @property
    def notes(self):
        return self._fs._feature_set.special_data.legal.notes

    @notes.setter
    def notes(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            legal_approved_notes=value,
            fields_to_update=[pb.FEATURE_SET_SPECIAL_DATA_LEGAL_NOTES],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.special_data.legal)

    def __str__(self):
        return f"Approved  : {self.approved} \n" f"Notes     : {self.notes} \n"


class Online:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._online = self._feature_set.online

    @property
    def online_namespace(self):
        return self._online.online_namespace

    @property
    def connection_type(self):
        return self._online.connection_type

    @property
    def topic(self):
        return self._online.topic

    def __repr__(self):
        return Utils.pretty_print_proto(self._online)

    def __str__(self):
        return (
            f"Online namespace  : {self.online_namespace} \n"
            f"Connection type   : {self.connection_type} \n"
            f"Topic             : {self.topic} \n"
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/FeatureSetProtoApi.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3ai/h2o/featurestore/api/v1/FeatureSetProtoApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"[\n\x10\x46\x65\x61tureSetHeader\x12\x12\n\nproject_id\x18\x01 \x01(\t\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x02 \x01(\t\x12\x1b\n\x13\x66\x65\x61ture_set_version\x18\x03 \x01(\t*)\n\x0e\x46\x65\x61tureSetType\x12\x07\n\x03RAW\x10\x00\x12\x0e\n\nENGINEERED\x10\x01*m\n\x0b\x46\x65\x61tureType\x12\x17\n\x13\x41UTOMATIC_DISCOVERY\x10\x00\x12\r\n\tCOMPOSITE\x10\x01\x12\r\n\tNUMERICAL\x10\x02\x12\x0f\n\x0b\x43\x41TEGORICAL\x10\x03\x12\x0c\n\x08TEMPORAL\x10\x04\x12\x08\n\x04TEXT\x10\x05\x42\x19\x42\x17\x46\x65\x61tureSetProtoApiProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'B\027FeatureSetProtoApiProto'
  _globals['_FEATURESETTYPE']._serialized_start=176
  _globals['_FEATURESETTYPE']._serialized_end=217
  _globals['_FEATURETYPE']._serialized_start=219
  _globals['_FEATURETYPE']._serialized_end=328
  _globals['_FEATURESETHEADER']._serialized_start=83
  _globals['_FEATURESETHEADER']._serialized_end=174
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/RecommendationProtoApi.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7ai/h2o/featurestore/api/v1/RecommendationProtoApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"L\n!RecommendationRegexMatchingPolicy\x12\r\n\x05regex\x18\x01 \x01(\t\x12\x18\n\x10percentage_match\x18\x02 \x01(\r\"\xbf\x01\n\"RecommendationSampleMatchingPolicy\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x02 \x01(\r\x12\x13\n\x0b\x63olumn_name\x18\x03 \x01(\t\x12\x17\n\x0fsample_fraction\x18\x04 \x01(\x01\x12\x16\n\x0e\x66uzzy_distance\x18\x05 \x01(\r\x12\x18\n\x10percentage_match\x18\x06 \x01(\r\"q\n$ListRecommendationClassifierResponse\x12I\n\x0b\x63lassifiers\x18\x01 \x03(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"\xdd\x01\n\x18RecommendationClassifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12N\n\x05regex\x18\x02 \x01(\x0b\x32=.ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicyH\x00\x12P\n\x06sample\x18\x03 \x01(\x0b\x32>.ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicyH\x00\x42\x11\n\x0fmatching_policy\"q\n%CreateRecommendationClassifierRequest\x12H\n\nclassifier\x18\x01 \x01(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"q\n%UpdateRecommendationClassifierRequest\x12H\n\nclassifier\x18\x01 \x01(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"@\n%DeleteRecommendationClassifierRequest\x12\x17\n\x0f\x63lassifier_name\x18\x01 \x01(\tB\x1d\x42\x1bRecommendationProtoApiProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'B\033RecommendationProtoApiProto'
  _globals['_RECOMMENDATIONREGEXMATCHINGPOLICY']._serialized_start=87
  _globals['_RECOMMENDATIONREGEXMATCHINGPOLICY']._serialized_end=163
  _globals['_RECOMMENDATIONSAMPLEMATCHINGPOLICY']._serialized_start=166
  _globals['_RECOMMENDATIONSAMPLEMATCHINGPOLICY']._serialized_end=357
  _globals['_LISTRECOMMENDATIONCLASSIFIERRESPONSE']._serialized_start=359
  _globals['_LISTRECOMMENDATIONCLASSIFIERRESPONSE']._serialized_end=472
  _globals['_RECOMMENDATIONCLASSIFIER']._serialized_start=475
  _globals['_RECOMMENDATIONCLASSIFIER']._serialized_end=696
  _globals['_CREATERECOMMENDATIONCLASSIFIERREQUEST']._serialized_start=698
  _globals['_CREATERECOMMENDATIONCLASSIFIERREQUEST']._serialized_end=811
  _globals['_UPDATERECOMMENDATIONCLASSIFIERREQUEST']._serialized_start=813
  _globals['_UPDATERECOMMENDATIONCLASSIFIERREQUEST']._serialized_end=926
  _globals['_DELETERECOMMENDATIONCLASSIFIERREQUEST']._serialized_start=928
  _globals['_DELETERECOMMENDATIONCLASSIFIERREQUEST']._serialized_end=992
# @@protoc_insertion_point(module_scope)



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .permission_base import PermissionBase
from .user import User


class Permission(PermissionBase):
    @property
    def user(self):
        return User(self._permission.user)


class ManageablePermission(Permission):
    @property
    def requestor(self):
        return User(self._permission.user)

    def revoke(self, reason):
        request = pb.RevokePermissionRequest(permission_id=self._permission.id, reason=reason)
        self._stub.RevokePermission(request)



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .permission_base import PermissionBase
from .user import User


class PermissionRequest(PermissionBase):
    @property
    def user(self):
        return User(self._permission.user)


class ManageablePermissionRequest(PermissionRequest):
    @property
    def requestor(self):
        return User(self._permission.user)

    def approve(self, reason):
        request = pb.ApprovePendingPermissionRequest(permission_id=self._permission.id, reason=reason)
        self._stub.ApprovePendingPermission(request)

    def reject(self, reason):
        request = pb.RejectPendingPermissionRequest(permission_id=self._permission.id, reason=reason)
        self._stub.RejectPendingPermission(request)



import datetime
import json

from google.protobuf.timestamp_pb2 import Timestamp

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from ..utils import Utils
from .ml_data_feature import MLDatasetFeature


class MLDataset:
    def __init__(self, stub: CoreServiceStub, ml_dataset_proto):
        ml_dataset = pb.MLDataset()
        ml_dataset.CopyFrom(ml_dataset_proto)
        self._ml_dataset = ml_dataset
        self._stub = stub

    @property
    def id(self):
        return self._ml_dataset.id

    @property
    def name(self):
        return self._ml_dataset.name

    @property
    def description(self):
        return self._ml_dataset.description

    @property
    def primary_keys(self):
        return self._ml_dataset.primary_keys

    @description.setter
    def description(self, value):
        update_request = pb.UpdateMLDatasetRequest(
            ml_dataset_id=self.id,
            description=value,
        )
        ml_dataset_proto = self._stub.UpdateMLDataset(update_request)
        ml_dataset = pb.MLDataset()
        ml_dataset.CopyFrom(ml_dataset_proto)
        self._ml_dataset = ml_dataset

    @property
    def start_date_time(self):
        timestamp: Timestamp = self._ml_dataset.start_date_time
        if timestamp:
            return datetime.datetime.fromtimestamp(timestamp.ToMilliseconds() / 1000, tz=datetime.timezone.utc)
        else:
            return None

    @property
    def end_date_time(self):
        timestamp: Timestamp = self._ml_dataset.end_date_time
        if timestamp:
            return datetime.datetime.fromtimestamp(timestamp.ToMilliseconds() / 1000, tz=datetime.timezone.utc)
        else:
            return None

    @property
    def features(self):
        return [MLDatasetFeature(feature) for feature in self._ml_dataset.features]

    def delete(self):
        request = pb.DeleteMLDatasetRequest(ml_dataset_id=self.id)
        self._stub.DeleteMLDataset(request)

    def download(self, output_dir=None):
        request = pb.RetrieveMLDatasetAsAsLinksRequest(ml_dataset_id=self.id)
        response = self._stub.RetrieveMLDatasetAsLinks(request)

        return Utils.download_files(output_dir, response.download_links)

    def as_spark_frame(self, spark_session):
        from ..commons.spark_utils import SparkUtils

        session_id = spark_session.conf.get("ai.h2o.featurestore.sessionId", "")
        request = pb.RetrieveMLDatasetAsSparkRequest(ml_dataset_id=self.id, session_id=session_id)

        response = self._stub.RetrieveMLDatasetAsSpark(request)
        spark_session.conf.set("ai.h2o.featurestore.sessionId", response.session_id)

        SparkUtils.configure_user_spark(spark_session)
        for k, v in response.options.items():
            spark_session.conf.set(k, v)

        return spark_session.read.format("parquet").load(response.path)

    def retrieve_online(self, *keys):
        request = pb.RetrieveMLDatasetOnlineRequest(ml_dataset_id=self.id, keys=[str(key) for key in keys])
        json_row = self._stub.RetrieveMLDatasetOnline(request).row
        return json.loads(json_row)

    def __repr__(self):
        return Utils.pretty_print_proto(self._ml_dataset)

    def __str__(self):
        nl = "\n"
        return (
            f"Name                  : {self.name} \n"
            f"Description           : {self.description} \n"
            f"Feature view version  : {self._ml_dataset.feature_view_version} \n"
            f"State                 : {self._ml_dataset.state} \n"
            f"Primary keys          : {self.primary_keys} \n"
            f"Features              : \n{nl.join(self._custom_feature_fields())} \n"
        )

    def _custom_feature_fields(self):
        tmp_list = list()
        for feature in self.features:
            s = (
                "   {\n"
                f"     Name             : {feature.name} \n"
                f"     Data type        : {feature.data_type} \n"
                f"     Is primary key?  : {feature.is_primary_key} \n"
                "   }"
            )
            tmp_list.append(s)
        return tmp_list



from ..base import ArrayField, BooleanField, StringField


class FeatureSet(object):
    name = StringField("featureSetName")
    description = StringField("description")
    deprecated = BooleanField("deprecated")
    application_name = StringField("applicationName")
    application_id = StringField("applicationId")
    tags = ArrayField("tags")
    data_source_domains = ArrayField("dataSourceDomains")
    owner = StringField("owner.name")
    author = StringField("author.name")
    features = ArrayField("features")



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..access_type import AccessType
from ..utils import Utils


class PermissionBase:
    def __init__(self, stub, permission, project_name=None):
        self._permission = permission
        self._stub = stub
        self._project_name = project_name

    @property
    def access_type(self):
        return AccessType.from_proto_permission(self._permission.permission)

    @property
    def status(self):
        return pb.PermissionState.Name(self._permission.state)

    @property
    def resource_id(self):
        return self._permission.resource_id

    @property
    def resource_type(self):
        return pb.ResourceType.Name(self._permission.resource_type)

    @property
    def reason(self):
        return self._permission.creation_reason

    @property
    def created_on(self):
        return Utils.timestamp_to_string(self._permission.created_on)

    @property
    def last_update_on(self):
        return Utils.timestamp_to_string(self._permission.last_update_on)

    def get_resource(self):
        if self._permission.resource_type == pb.ResourceType.PROJECT:
            request = pb.GetProjectRequest(project_name=self._project_name)
            response = self._stub.GetProject(request)
            return response.project
        elif self._permission.resource_type == pb.ResourceType.FEATURE_SET:
            request = pb.FeatureSetRequestId(feature_set_id=self._permission.resource_id)
            response = self._stub.GetFeatureSetById(request)
            return response.feature_set
        else:
            raise Exception("Unknown resource type")

    def __repr__(self):
        return Utils.pretty_print_proto(self._permission)



from .base_job import BaseJob
from .ml_dataset import MLDataset


class CreateMLDatasetJob(BaseJob):
    def _response_method(self, job_id):
        return MLDataset(self._stub, self._stub.GetMLDatasetCreationJobOutput(job_id))



class FeatureSetRef:
    def __init__(self, id: str, major_version: int):
        self.id = id
        self.major_version = major_version

    def __repr__(self):
        return f"FeatureSetRef({self.id}, {self.major_version})"

    def __hash__(self):
        return hash((self.id, self.major_version))

    def __eq__(self, other):
        return self.id == other.id and self.major_version == other.major_version



from abc import ABC, abstractmethod

from google.protobuf.empty_pb2 import Empty

import ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2 as pb


class Classifier(ABC):
    @abstractmethod
    def _to_proto(self):
        raise NotImplementedError("Method `__to_proto` needs to be implemented by the child class")

    @staticmethod
    def _from_proto(classifier: pb.RecommendationClassifier):
        oneof_kind = classifier.WhichOneof("matching_policy")
        if oneof_kind == "regex":
            return RegexClassifier(
                classifier.name,
                classifier.regex.regex,
                classifier.regex.percentage_match,
            )
        elif oneof_kind == "sample":
            return SampleClassifier(
                classifier.name,
                classifier.sample.feature_set_id,
                classifier.sample.feature_set_major_version,
                classifier.sample.column_name,
                classifier.sample.sample_fraction,
                classifier.sample.fuzzy_distance,
                classifier.sample.percentage_match,
            )
        elif oneof_kind is None:
            return EmptyClassifier(classifier.name)
        else:
            raise ValueError("Not supported classifier provided")


class EmptyClassifier(Classifier):
    def __init__(self, name):
        self.name = name

    def _to_proto(self):
        return pb.RecommendationClassifier(name=self.name)

    def __repr__(self):
        return f"EmptyClassifier(name={self.name})"


class RegexClassifier(Classifier):
    def __init__(self, name, regex, percentage_match):
        self.name = name
        self.regex = regex
        self.percentage_match = percentage_match

    def _to_proto(self):
        return pb.RecommendationClassifier(
            name=self.name,
            regex=pb.RecommendationRegexMatchingPolicy(regex=self.regex, percentage_match=self.percentage_match),
        )

    def __repr__(self):
        return f"RegexClassifier(name={self.name}, regex={self.regex}, percentage_match={self.percentage_match})"


class SampleClassifier(Classifier):
    def __init__(
        self,
        name: str,
        feature_set_id: str,
        feature_set_major_version: int,
        column_name: str,
        sample_fraction: float,
        fuzzy_distance: int,
        percentage_match: int,
    ):
        self.name = name
        self.feature_set_id = feature_set_id
        self.feature_set_major_version = feature_set_major_version
        self.column_name = column_name
        self.sample_fraction = sample_fraction
        self.fuzzy_distance = fuzzy_distance
        self.percentage_match = percentage_match

    @classmethod
    def from_feature_set(
        cls,
        feature_set,
        name,
        column_name,
        sample_fraction,
        fuzzy_distance,
        percentage_match,
    ):
        return cls(
            name,
            feature_set.id,
            feature_set.major_version,
            column_name,
            sample_fraction,
            fuzzy_distance,
            percentage_match,
        )

    def _to_proto(self):
        return pb.RecommendationClassifier(
            name=self.name,
            sample=pb.RecommendationSampleMatchingPolicy(
                feature_set_id=self.feature_set_id,
                feature_set_major_version=self.feature_set_major_version,
                column_name=self.column_name,
                sample_fraction=self.sample_fraction,
                fuzzy_distance=self.fuzzy_distance,
                percentage_match=self.percentage_match,
            ),
        )

    def __repr__(self):
        return (
            f"SampleClassifier(name={self.name}, "
            f"feature_set_id={self.feature_set_id}, "
            f"feature_set_major_version={self.feature_set_major_version}, "
            f"column_name={self.column_name}, "
            f"sample_fraction={self.sample_fraction}, "
            f"fuzzy_distance={self.fuzzy_distance}, "
            f"percentage_match={self.percentage_match})"
        )


class Classifiers:
    def __init__(self, stub):
        self._stub = stub

    def list(self) -> [Classifier]:
        request = Empty()
        response = self._stub.ListRecommendationClassifiers(request)
        return [Classifier._from_proto(classifier) for classifier in response.classifiers]

    def create(self, classifier) -> None:
        if isinstance(classifier, str):
            classifier_to_send = EmptyClassifier(classifier)
        elif isinstance(classifier, Classifier):
            classifier_to_send = classifier
        else:
            raise ValueError("Parameter classifier should be string or object of Classifier class")

        request = pb.CreateRecommendationClassifierRequest(classifier=classifier_to_send._to_proto())
        self._stub.CreateRecommendationClassifier(request)

    def update(self, classifier: Classifier):
        request = pb.UpdateRecommendationClassifierRequest(classifier=classifier._to_proto())
        self._stub.UpdateRecommendationClassifier(request)

    def delete(self, name: str):
        request = pb.DeleteRecommendationClassifierRequest(classifier_name=name)
        self._stub.DeleteRecommendationClassifier(request)

    def __repr__(self):
        return "Recommendation classifiers"



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='h2o-featurestore',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.12.8",
    description='Feature Store Client for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://docs.h2o.ai/feature-store/latest-stable/docs/index.html',
    download_url='https://docs.h2o.ai/feature-store/latest-stable/docs/download.html',
    author='H2O.ai',
    author_email='support@h2o.ai',
    license='Apache v2',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8"
    ],
    keywords='machine learning, data mining, statistical analysis, modeling, big data, distributed, parallel',

    # find python packages starting in the current directory
    packages=find_packages(),

    # run-time dependencies
    install_requires=[
        "grpcio>=1.30.0",
        "protobuf>=3.19.0",
        "jproperties==2.1.1",
        "requests",
        "python-dateutil",
        "grpcio-tools>=1.30.0",
        "googleapis-common-protos>=1.57.0"
    ],

    # bundled binary packages
    package_data={
        'featurestore': ['version.txt', '_credentials/roots.pem']},
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/Events.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'ai/h2o/featurestore/api/v1/Events.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"o\n\x11ModificationEvent\x12Q\n\x13\x66\x65\x61ture_set_deleted\x18\x01 \x01(\x0b\x32\x32.ai.h2o.featurestore.api.v1.FeatureSetDeletedEventH\x00\x42\x07\n\x05value\"0\n\x16\x46\x65\x61tureSetDeletedEvent\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\tB\rB\x0b\x45ventsProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.Events_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'B\013EventsProto'
  _globals['_MODIFICATIONEVENT']._serialized_start=71
  _globals['_MODIFICATIONEVENT']._serialized_end=182
  _globals['_FEATURESETDELETEDEVENT']._serialized_start=184
  _globals['_FEATURESETDELETEDEVENT']._serialized_end=232
# @@protoc_insertion_point(module_scope)
